// Runtime names and paths

#define _GNU_SOURCE

#include "runtime.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

static _Bool init_failed;
static char *prog_name;
static char *root_dir;
static _Bool usr_dir;

static char *remove_prefix(char *path, char *end);

__attribute__((constructor)) static void unit_constructor(void) {
	char *path = canonicalize_file_name("/proc/self/exe");
	if (!path) {
		init_failed = 1;
		return;
	}

	char *slash = strrchr(path, '/');
	if (!slash) {
		init_failed = 1;
		goto cleanup;
	}
	*slash = '\0';

	prog_name = strdup(slash + 1);
	if (!prog_name) {
		init_failed = 1;
		goto cleanup;
	}

	char *new_end = remove_prefix(path, slash);
	if (new_end) {
		size_t path_len = new_end - path;
		const char *prefix = "/usr";
		if (path_len >= strlen(prefix)) {
			char *p = new_end - strlen(prefix);
			if (!strcmp(p, prefix)) {
				*p = '\0';
				usr_dir = 1;
			}
		}
	}

	if (path[0]) {
		root_dir = strdup(path);
		if (!root_dir)
			init_failed = 1;
	}
cleanup:
	free(path);
}

__attribute__((destructor)) static void unit_destructor(void) {
	free(prog_name);
	free(root_dir);
}

static char *remove_prefix(char *path, char *end) {
	size_t path_len = end - path;

	static const char *prefixes[] = { "/bin", "/games" };
	int i;
	for (i = 0; i < 2; i++) {
		const char *prefix = prefixes[i];
		if (path_len >= strlen(prefix)) {
			char *p = end - strlen(prefix);
			if (!strcmp(p, prefix)) {
				*p = '\0';
				return p;
			}
		}
	}

	static const char LIB[] = "/lib/";
	if (path_len >= strlen(LIB) + strlen(prog_name)) {
		char *p = end - strlen(prog_name);
		if (!strcmp(p, prog_name)) {
			p -= strlen(LIB);
			if (!strncmp(p, LIB, strlen(LIB))) {
				*p = '\0';
				return p;
			}
		}
	}
	return NULL;
}

char *runtime_etc_dir(void) {
	if (init_failed)
		return NULL;
	char *dir;
	if (asprintf(&dir, "%s/etc/%s", root_dir ? : "", prog_name) == -1)
		return NULL;
	return dir;
}

_Bool runtime_init_failed(void) {
	return init_failed;
}

const char *runtime_prog_name(void) {
	return prog_name;
}

char *runtime_run_dir(void) {
	if (init_failed)
		return NULL;
	char *dir;
	if (asprintf(&dir, "%s/var/run/%s", root_dir ? : "", prog_name) == -1)
		return NULL;
	return dir;
}

char *runtime_share_dir(void) {
	if (init_failed)
		return NULL;
	char *dir;
	if (asprintf(&dir, "%s%s/share/%s", root_dir ? : "",
		usr_dir ? "/usr" : "", prog_name) == -1)
		return NULL;
	return dir;
}
