// UTF-8 handling

#include "utf8.h"

static void utf8_length_action(unsigned code, void *data);
static void utf8_valid_action(unsigned code, void *data);

size_t utf8_length(const char *s, _Bool *error) {
	size_t length = 0;
	*error = utf8_parse(s, utf8_length_action, &length);
	return length;
}

static void utf8_length_action(__attribute__((unused)) unsigned code,
	void *data) {
	size_t *length = data;
	(*length)++;
}

_Bool utf8_parse(const char *s, utf8_action action, void *data) {
	static const unsigned min_code[] = { 0x00000000, 0x00000080, 0x00000800,
		0x00010000, 0x00200000, 0x04000000 };
	while (*s) {
		unsigned code;
		if (*s & 0x80) {
			if (!(*s & 0x40))
				return 1;
			int class = 1;
			char bit_mask = 0x20;
			while (*s & bit_mask) {
				if (bit_mask == 0x02)
					return 1;
				class++;
				bit_mask >>= 1;
			}
			code = *s++ & (0x3f >> class);
			int i;
			for (i = 0; i < class; i++) {
				if ((*s & 0xc0) != 0x80)
					return 1;
				code <<= 6;
				code |= *s++ & 0x3f;
			}
			if (code < min_code[class])
				return 1;
		}
		else
			code = *s++;
		action(code, data);
	}
	return 0;
}

_Bool utf8_valid(const char *s) {
	return !utf8_parse(s, utf8_valid_action, NULL);
}

static void utf8_valid_action(__attribute__((unused)) unsigned code,
	__attribute__((unused)) void *data) {
}
